@@ -17,6 +17,7 @@ import pymaid
 import navis
 import requests
 import textwrap
+import time
 
 import datetime as dt
 import numpy as np
@@ -26,11 +27,13 @@ from concurrent import futures
 from diskcache import Cache
 from requests_futures.sessions import FuturesSession
 from scipy import ndimage
+from tqdm.auto import trange
 
 from .. import spine
 from .. import xform
 
-from .utils import parse_volume, FLYWIRE_DATASETS, get_chunkedgraph_secret
+from .utils import (parse_volume, FLYWIRE_DATASETS, get_chunkedgraph_secret,
+                    retry, get_cave_client)
 
 
 __all__ = ['fetch_edit_history', 'fetch_leaderboard', 'locs_to_segments',
@@ -255,12 +258,13 @@ def roots_to_supervoxels(x, use_cache=True, dataset='production', progress=True)
     # mismatch in types (int vs np.int?) which causes all root IDs to be in miss
     # -> I think that's because of the way disk cache works
     miss = x[~np.isin(x, np.array(list(svoxels.keys())).astype(int))]
-    svoxels.update({i: vol.get_leaves(i,
-                                      bbox=vol.meta.bounds(0),
-                                      mip=0) for i in navis.config.tqdm(miss,
-                                                           desc='Querying',
-                                                           disable=not progress,
-                                                           leave=False)})
+    get_leaves = retry(vol.get_leaves)
+    svoxels.update({i: get_leaves(i,
+                                  bbox=vol.meta.bounds(0),
+                                  mip=0) for i in navis.config.tqdm(miss,
+                                                                    desc='Querying',
+                                                                    disable=not progress,
+                                                                    leave=False)})
 
     # Update cache
     if use_cache:
@@ -271,22 +275,25 @@ def roots_to_supervoxels(x, use_cache=True, dataset='production', progress=True)
     return svoxels
 
 
-def supervoxels_to_roots(x, use_cache=False, dataset='production'):
+def supervoxels_to_roots(x, timestamp=None, batch_size=10_000,
+                         retry=True, progress=True, dataset='production'):
     """Get root(s) for given supervoxel(s).
 
     Parameters
     ----------
     x :             int | list of int
                     Supervoxel ID(s) to find the root(s) for.
-    use_cache :     bool
-                    Whether to use disk cache to avoid repeated queries for the
-                    same supervoxel. The implementation for this is optimized
-                    for a small memory footprint - so it's slow! At this point
-                    only use it if you want to minimize your impact on the
-                    flywire backends and you have time on your hands. The cache
-                    is stored in `~/.fafbseg/`.
+    timestamp :     int | str | datetime
+                    Get roots at given date (and time). Int must be unix
+                    timestamp. String must be ISO 8601 - e.g. '2021-11-15'.
+                    If timestamp is given, will ignore `use_cache`.
+    batch_size :    int
+                    Max number of supervoxel IDs per query. Reduce batch size if
+                    you experience time outs.
+    retry :         bool
+                    Whether to retry if a batched query fails.
     dataset :       str | CloudVolume
-                    Against which flywire dataset to query::
+                    Against which dataset to query::
                         - "production" (current production dataset, fly_v31)
                         - "sandbox" (i.e. fly_v26)
     progress :      bool
@@ -294,8 +301,8 @@ def supervoxels_to_roots(x, use_cache=False, dataset='production'):
 
     Returns
     -------
-    dict
-                    ``{root_id: [svoxel_id1, svoxelid2, ...], ...}``
+    roots  :        numpy array
+                    Roots corresponding to supervoxels in `x`.
 
     Examples
     --------
@@ -312,49 +319,32 @@ def supervoxels_to_roots(x, use_cache=False, dataset='production'):
     # Prepare results array
     roots = np.zeros(x.shape, dtype=np.int64)
 
-    # We can't query supervoxel ID 0
-    not_zero = x != 0
+    if isinstance(timestamp. np.datetime64):
+        timestamp = str(timestamp)
 
-    if use_cache:
-        # Cache for supervoxel -> root map
-        # Grows to max 1Gb by default and persists across sessions
-        with Cache(directory='~/.fafbseg/roots_cache/') as roots_cache:
-            # See if we have any of these supervoxels cached
-            with roots_cache.transact():
-                is_cached = np.isin(x, roots_cache)
+    for i in trange(0, len(x), int(batch_size),
+                    desc='Fetching roots',
+                    disable=not progress or len(x) < batch_size):
+        # This batch
+        batch = x[i:i+batch_size]
 
-            # For cached roots, check if they are latest
-            if np.any(is_cached):
-                # Get values from cache
-                with roots_cache.transact():
-                    cached = np.array([roots_cache[i] for i in x[is_cached]])
-                # Check if cached roots are latest
-                is_latest = is_latest_root(cached, dataset=dataset)
-                # Set roots that are still up-to-date
-                is_cached[is_cached] = is_latest
-                roots[is_cached] = cached[is_latest]
-
-            # To fetch are those supervoxels that are != 0 and are not cached
-            to_fetch = ~is_cached & not_zero
-
-            # Fill in the blanks
-            if np.any(to_fetch):
-                roots[to_fetch] = vol.get_roots(x[to_fetch])
-
-                # Update cache
-                with roots_cache.transact():
-                    for k, v in zip(x[to_fetch], roots[to_fetch]):
-                        roots_cache[k] = v
-
-    else:
         # get_roots() doesn't like to be asked for zeros - causes server error
-        roots[not_zero] = vol.get_roots(x[not_zero])
+        not_zero = batch != 0
+        try:
+            roots[i:i+batch_size][not_zero] = vol.get_roots(batch[not_zero],
+                                                            timestamp=timestamp)
+        except BaseException:
+            if not retry:
+                raise
+            time.sleep(1)
+            roots[i:i+batch_size][not_zero] = vol.get_roots(batch[not_zero],
+                                                            timestamp=timestamp)
 
     return roots
 
 
 def locs_to_supervoxels(locs, mip=2, coordinates='voxel'):
-    """Retrieve flywire supervoxel IDs at given location(s).
+    """Retrieve FlyWire supervoxel IDs at given location(s).
 
     Use Eric Perlman's service on spine.
 
@@ -460,7 +450,7 @@ def neuron_to_segments(x, dataset='production', coordinates='voxel'):
     return matrix
 
 
-def locs_to_segments(locs, root_ids=True, dataset='production',
+def locs_to_segments(locs, root_ids=True, timestamp=None, dataset='production',
                      coordinates='voxel'):
     """Retrieve flywire segment IDs (root IDs) at given location(s).
 
@@ -473,6 +463,10 @@ def locs_to_segments(locs, root_ids=True, dataset='production',
     root_ids :      bool
                     If True, will return root IDs. If False, will return supervoxel
                     IDs.
+    timestamp :     int | str | datetime, optional
+                    Get roots at given date (and time). Int must be unix
+                    timestamp. String must be ISO 8601 - e.g. '2021-11-15'.
+                    If timestamp is given, will ignore `use_cache`.
     dataset :       str | CloudVolume
                     Against which flywire dataset to query::
                         - "production" (current production dataset, fly_v31)
@@ -501,7 +495,7 @@ def locs_to_segments(locs, root_ids=True, dataset='production',
     if not root_ids:
         return svoxels
 
-    return supervoxels_to_roots(svoxels, dataset=dataset)
+    return supervoxels_to_roots(svoxels, timestamp=timestamp, dataset=dataset)
 
 
 def skid_to_id(x,
@@ -628,17 +622,24 @@ def is_latest_root(id, dataset='production', **kwargs):
 
     id = navis.utils.make_iterable(id).astype(str)
 
+    # The server doesn't like being asked for zeros
+    not_zero = id != '0'
+
+    is_latest = np.ones(len(id)).astype(bool)
+
     session = requests.Session()
     token = get_chunkedgraph_secret()
     session.headers['Authorization'] = f"Bearer {token}"
 
     url = f'https://prod.flywire-daf.com/segmentation/api/v1/table/{dataset}/is_latest_roots?int64_as_str=1'
-    post = {'node_ids': id.tolist()}
+    post = {'node_ids': id[not_zero].tolist()}
     r = session.post(url, json=post)
 
     r.raise_for_status()
 
-    return np.array(r.json()['is_latest'])
+    is_latest[not_zero] = np.array(r.json()['is_latest'])
+
+    return is_latest
 
 
 def update_ids(id,
@@ -648,11 +649,13 @@ def update_ids(id,
     """Retrieve the most recent version of given FlyWire (root) ID(s).
 
     This function works by:
-        1. Checking if ID is outdated (see fafbseg.flywire.is_latest_root)
-        2. Fetching all supervoxels for outdated IDs
-        3. Picking a random sample of ``sample`` of these supervoxels
+        1. Check if ID is outdated (see fafbseg.flywire.is_latest_root)
+        2. See if we can map outdated IDs to a single up-to-date root (works
+           if neuron has only seen merges)
+        3. For uncertain IDs, fetch all supervoxels
+        3. Picking a random sample (see ``sample`` parameter) of supervoxels
         4. Fetching the most recent root IDs for the sample supervoxels
-        5. Returning the root ID that was hit the most.
+        5. Returning the root ID that was hit the most
 
     Parameters
     ----------
@@ -698,15 +701,16 @@ def update_ids(id,
     vol = parse_volume(dataset, **kwargs)
 
     if isinstance(id, (list, set, np.ndarray)):
+        # Run is_latest once for all roots
         is_latest = is_latest_root(id, dataset=dataset)
         res = [update_ids(x,
-                          dataset=vol,
+                          dataset=dataset,
                           is_latest=il,
                           sample=sample) for x, il in navis.config.tqdm(zip(id, is_latest),
-                                                           desc='Updating',
-                                                           leave=False,
-                                                           total=len(id),
-                                                           disable=not progress or len(id) == 1)]
+                                                                        desc='Updating',
+                                                                        leave=False,
+                                                                        total=len(id),
+                                                                        disable=not progress or len(id) == 1)]
         return pd.concat(res, axis=0, sort=False, ignore_index=True)
 
     # Check if outdated
@@ -714,37 +718,45 @@ def update_ids(id,
         is_latest = is_latest_root(id, dataset=dataset)[0]
 
     if not is_latest:
-        # Get supervoxel ids - we need to use mip=0 because otherwise small neurons
-        # might not have any (visible) supervoxels
-        svoxels = vol.get_leaves(id, bbox=vol.meta.bounds(0), mip=0)
+        client = get_cave_client(dataset=dataset)
+        pot_roots = client.chunkedgraph.get_latest_roots(id)
 
-        # Shuffle voxels
-        np.random.shuffle(svoxels)
-
-        # Generate sample
-        if sample >= 1:
-            smpl = svoxels[: sample]
-        else:
-            smpl = svoxels[: int(len(svoxels) * sample)]
-
-        # Fetch up-to-date root IDs for the sampled supervoxels
-        roots = supervoxels_to_roots(smpl, dataset=vol)
-
-        # Find unique Ids and count them
-        unique, counts = np.unique(roots, return_counts=True)
-
-        # Get sorted indices
-        sort_ix = np.argsort(counts)
-
-        # New Id is the most frequent ID
-        new_id = unique[sort_ix[-1]]
-
-        # Confidence is the difference between the top and the 2nd most frequent ID
-        if len(unique) > 1:
-            conf = round((counts[sort_ix[-1]] - counts[sort_ix[-2]]) / sum(counts),
-                         2)
-        else:
+        if len(pot_roots) == 1:
+            new_id = pot_roots[0]
             conf = 1
+        else:
+            # Get supervoxel ids - we need to use mip=0 because otherwise small
+            # neurons might not have any (visible) supervoxels
+            svoxels = roots_to_supervoxels(id, progress=False)[int(id)]
+
+            # Shuffle voxels
+            np.random.shuffle(svoxels)
+
+            # Generate sample
+            if sample >= 1:
+                smpl = svoxels[: sample]
+            else:
+                smpl = svoxels[: int(len(svoxels) * sample)]
+
+            # Fetch up-to-date root IDs for the sampled supervoxels
+            roots = supervoxels_to_roots(smpl, dataset=vol, progress=False)
+
+            # Find unique Ids and count them
+            unique, counts = np.unique(roots, return_counts=True)
+
+            # Get sorted indices
+            sort_ix = np.argsort(counts)
+
+            # New Id is the most frequent ID
+            new_id = unique[sort_ix[-1]]
+
+            # Confidence is the difference between the top and the 2nd most
+            # frequent ID
+            if len(unique) > 1:
+                conf = round((counts[sort_ix[-1]] - counts[sort_ix[-2]]) / sum(counts),
+                             2)
+            else:
+                conf = 1
     else:
         new_id = id
         conf = 1
