@@ -20,10 +20,11 @@ import numpy as np
 import pandas as pd
 
 from functools import partial
+from pathlib import Path
 from tqdm.auto import trange
 
 from .segmentation import is_latest_root
-from .utils import parse_root_ids, get_cave_client
+from .utils import parse_root_ids, get_cave_client, retry, get_synapse_areas
 
 from ..synapses.utils import catmaid_table
 from ..synapses.transmitters import collapse_nt_predictions
@@ -110,15 +111,15 @@ def predict_transmitter(x, single_pred=False, weighted=True, live_query=True,
                                    weighted=weighted, id_col='pre')
 
 
-def fetch_synapses(x, pre=True, post=True, attach=True, min_score=0, clean=True,
-                   transmitters=False, live_query=True, batch_size=100,
-                   dataset='production', progress=True):
+def fetch_synapses(x, pre=True, post=True, attach=True, min_score=30, clean=True,
+                   transmitters=False, neuropils=False, live_query=True,
+                   batch_size=100, dataset='production', progress=True):
     """Fetch Buhmann et al. (2019) synapses for given neuron(s).
 
     Parameters
     ----------
     x :             int | list of int | Neuron/List
-                    Either a flywire segment ID (i.e. root ID), a list thereof
+                    Either a FlyWire segment ID (i.e. root ID), a list thereof
                     or a Neuron/List. For neurons, the ``.id`` is assumed to be
                     the root ID. If you have a neuron (in FlyWire space) but
                     don't know its ID, use
@@ -130,11 +131,14 @@ def fetch_synapses(x, pre=True, post=True, attach=True, min_score=0, clean=True,
     transmitters :  bool
                     Whether to also load per-synapse neurotransmitter predictions
                     from Eckstein et al. (2020).
+    neuropils :     bool
+                    Whether to add a column indicating which neuropil a synapse
+                    is in.
     attach :        bool
                     If True and ``x`` is Neuron/List, the synapses will be added
                     as ``.connectors`` table. For TreeNeurons (skeletons), the
                     synapses will be mapped to the closest node.
-    min_score :     int
+    min_score :     int, optional
                     Minimum "cleft score". The default of 30 is what Buhmann et al.
                     used in the paper.
     clean :         bool
@@ -158,6 +162,11 @@ def fetch_synapses(x, pre=True, post=True, attach=True, min_score=0, clean=True,
     Returns
     -------
     pandas.DataFrame
+                    Note that each synapse (or rather synaptic connection)
+                    will show up only once. Depending on the query neurons
+                    (`x`), a given row might represent a presynapse for one and
+                    a postsynapse for another neuron.
+
 
     """
     if not pre and not post:
@@ -176,19 +185,19 @@ def fetch_synapses(x, pre=True, post=True, attach=True, min_score=0, clean=True,
     client = get_cave_client(dataset=dataset)
 
     columns = ['pre_pt_root_id', 'post_pt_root_id', 'cleft_score',
-               'pre_pt_position', 'post_pt_position']
+               'pre_pt_position', 'post_pt_position', 'id']
 
     if transmitters:
         columns += ['gaba', 'ach', 'glut', 'oct', 'ser', 'da']
 
     if live_query:
-        func = partial(client.materialize.live_query,
+        func = partial(retry(client.materialize.live_query),
                        table=client.materialize.synapse_table,
                        timestamp=dt.datetime.utcnow(),
                        split_positions=True,
                        select_columns=columns)
     else:
-        func = partial(client.materialize.query_table,
+        func = partial(retry(client.materialize.query_table),
                        table=client.materialize.synapse_table,
                        split_positions=True,
                        select_columns=columns)
@@ -206,6 +215,9 @@ def fetch_synapses(x, pre=True, post=True, attach=True, min_score=0, clean=True,
     # Combine results from batches
     syn = pd.concat(syn, axis=0, ignore_index=True)
 
+    # Depending on how queries were batched, we need to drop duplicate synapses
+    syn.drop_duplicates('id', inplace=True)
+
     # Rename some of those columns
     syn.rename({'post_pt_root_id': 'post',
                 'pre_pt_root_id': 'pre',
@@ -240,25 +252,32 @@ def fetch_synapses(x, pre=True, post=True, attach=True, min_score=0, clean=True,
     # Avoid copy warning
     syn = syn.copy()
 
+    if neuropils:
+        syn['neuropil'] = get_synapse_areas(syn['id'].values)
+        syn['neuropil'] = syn.neuropil.astype('category')
+
+    # Drop ID column
+    # syn.drop('id', axis=1, inplace=True)
+
     if attach and isinstance(x, navis.NeuronList):
         for n in x:
             presyn = postsyn = pd.DataFrame([])
             if pre:
                 presyn = syn.loc[syn.pre == int(n.id),
                                  ['pre_x', 'pre_y', 'pre_z',
-                                  'cleft_scores', 'post']].rename({'pre_x': 'x',
-                                                                   'pre_y': 'y',
-                                                                   'pre_z': 'z',
-                                                                   'post': 'partner_id'},
+                                  'cleft_score', 'post']].rename({'pre_x': 'x',
+                                                                  'pre_y': 'y',
+                                                                  'pre_z': 'z',
+                                                                  'post': 'partner_id'},
                                                                   axis=1)
                 presyn['type'] = 'pre'
             if post:
                 postsyn = syn.loc[syn.post == int(n.id),
                                   ['post_x', 'post_y', 'post_z',
-                                   'cleft_scores', 'pre']].rename({'post_x': 'x',
-                                                                   'post_y': 'y',
-                                                                   'post_z': 'z',
-                                                                   'pre': 'partner_id'},
+                                   'cleft_score', 'pre']].rename({'post_x': 'x',
+                                                                  'post_y': 'y',
+                                                                  'post_z': 'z',
+                                                                  'pre': 'partner_id'},
                                                                   axis=1)
                 postsyn['type'] = 'post'
 
@@ -340,12 +359,12 @@ def fetch_adjacency(sources, targets=None, min_score=30, live_query=True,
 
     columns = ['pre_pt_root_id', 'post_pt_root_id', 'cleft_score']
     if live_query:
-        func = partial(client.materialize.live_query,
+        func = partial(retry(client.materialize.live_query),
                        table=client.materialize.synapse_table,
                        timestamp=dt.datetime.utcnow(),
                        select_columns=columns)
     else:
-        func = partial(client.materialize.query_table,
+        func = partial(retry(client.materialize.query_table),
                        table=client.materialize.synapse_table,
                        select_columns=columns)
 
@@ -467,12 +486,12 @@ def fetch_connectivity(x, clean=True, style='simple', min_score=30,
         columns += ['gaba', 'ach', 'glut', 'oct', 'ser', 'da']
 
     if live_query:
-        func = partial(client.materialize.live_query,
+        func = partial(retry(client.materialize.live_query),
                        table=client.materialize.synapse_table,
                        timestamp=dt.datetime.utcnow(),
                        select_columns=columns)
     else:
-        func = partial(client.materialize.query_table,
+        func = partial(retry(client.materialize.query_table),
                        table=client.materialize.synapse_table,
                        select_columns=columns)
 
